* ERRHDEMO.PRG: wrapper UDF for form to run a VFP command line with error-handling
* By M. Asherman, Copyright (c) 1999, SpaceTime Systems
* 11/20/99: added this wrapper function to make form testing easier.
* 11/21/99: renamed property ehfp_returnval to ehfp_success, to clarify meaning.
* 11/24/99: adjust for renamed ehfp_... -> ehp_...
* 11/26/99: fix logic to avoid bombing on invalid Action arg and
*	eliminate use of WAIT WINDOW error reporting no longer needed.
* 11/27/00: introduce optional READ EVENTS to force main program depth > 0, which 
*	is necessary for proper catch-all error handling logic because of limitations
*	in VFP's RETURN TO ... command (see task erma0005).
* 12/11/00: renamed demo-related components to avoid confusion with generic.
* 05/29/02: beef up logic to avoid errors when target error handler disappears,
*	after changes for support of a global error handler and main Demo menu. 
* 05/31/02: add extra menu cleanup logic to deal with modal Abort cases.
*	Also add logic to prevent creation of multiple demoform instance, since this
*	demo is designed to support only one instance.  Eliminate obsolete ERRH_PUBNM.
*	Also fix Run logic to coerce command line argument to character type.
*
* Note that this wrapper was coded in such a way as to avoid depending on 
* "target indirection", except for modeless cases that use READ EVENTS, 
* where there is no other way to obtain a meaningful result.

function errhdemo
* Returns a success flag result, which is set to .t. only if completely successful.

* takes 3 optional arguments
lparameters action_arg, cmdlin_arg, noread_arg
* action_arg:	one of the reserved action codes: 'Prompt' (the default), 'PromptN', 
*				or 'Run', 'RunN'.  
*				Brief descriptions and recommended macros for these codes are in ErrHandl.h.
*				The default is 'Prompt' (ERRH_ACPRM), which means to use a modal dialog for 
*				interactive operation.
* cmdlin_arg:	optional initial command line for interactive cases,
*				or the command line to be run non-interactively.
* noread_arg:	optional flag that suppresses use of READ EVENTS in modeless cases.
*				READ EVENTS is used by default for modeless invocations,
*				in order to assure proper handling of abnormal errors, and this
*				would also be required under runtime VFP.

#include errhandl.h				&& macro definitions for MDA Error Handler facilities

local demoobjref, cmdline, act, result, targname

* check for existing errhdemo form
if vartype(m.ERRH_MVGLOBAL) = 'O' ;
  and vartype(m.ERRH_MVGLOBAL.errhdp_demoform) # 'U'	&& the demo has already been invoked
	demoobjref = m.ERRH_MVGLOBAL.errhdp_demoform		&& reference to demoform or NULL if none
endif

* coerce command line argument to required character type
cmdline = iif(vartype(m.cmdlin_arg) = 'C', m.cmdlin_arg, '')

* Sort out various Action cases that bear on how we invoke the form and how we obtain results.
* Carefully process 1st argument, Action code, to avoid encountering errors,
* because we haven't yet established full error-handling initializations.
act = iif(empty(m.action_arg), ERRH_ACPRM, ;
  iif(vartype(m.action_arg) = 'C', lower(trim(m.action_arg)), ''))	&& default = Prompt
do case											&& dispatch on Action argument
  case m.act == ERRH_ACPRN						&& Prompt with Non-modal (modeless) dialog
	* check for the possibility that there is already an instance of demoform running
	if vartype(m.demoobjref) = 'O' 				&& demoform is already running
		m.demoobjref.text1.value = m.cmdline	&& assign command to execute
	else										&& the errhdemo dialog is not yet present
		* Launch modeless dialog form that will persist beyond this wrapper function call
		do form errhdemo NAME demoobjref with m.action_arg, m.cmdlin_arg 	&& pass along same args
	endif

  case m.act == ERRH_ACRNN						&& Run non-interactively with Non-modal dialog
	* see if reserved object already exists, so we can reuse it without re-creating
	if vartype(m.demoobjref) = 'O' 				&& demoform is already running
		* prelim logic. - ...Run method really should take a command line arg and return flag
		m.demoobjref.text1.value = m.cmdline	&& assign command to execute
		m.demoobjref.ErrHCM_Run()				&& invoke the form's Run method
	else										&& we'll need to instantiate
		* Invoke modeless dialog to run command non-interactively in hidden form 
		* that will persist beyond this wrapper function call.
		do form errhdemo NAME demoobjref with m.action_arg, m.cmdlin_arg	&& pass same args
	endif
	
  otherwise										&& Modal (synchronous) invocation cases
	* check for the possibility that there is already an instance of demoform running
	if vartype(m.demoobjref) = 'O' 				&& an instance of demoform is already running
		m.demoobjref.release()					&& discard the previous demoform instance
	endif
	
	* launch modal dialog form that does the work and returns success flag directly
	do form errhdemo TO m.result with m.action_arg, m.cmdlin_arg		&& pass along same args

	* check for main menu residue that may not have been cleaned up properly after an Abort
	if vartype(m.ERRH_MVGLOBAL) # 'O'			&& there is no global error handler
		* Perform cleanups that really should have been done by the global handler,
		* but which may not have worked because of the presence of a modal dialog.
		RELEASE PAD errhdempad OF _msysmenu		&& remove the Demo pad from main menu, if still there
		release popups errhdempop extended		&& clean up the Demo menu popup, if still allocated
	endif
	
	* finish up for modal cases
	if vartype(m.result) = 'L'					&& result is well-defined
		return m.result							&& done
	endif
	return .f.									&& fail
endcase

* modeless cases - check for unexpected failure to launch the demo form
if vartype(m.demoobjref) # 'O'					&& demo form object ref not available
	return .f.									&& fail - assume error has already been reported
endif

* Check for deliberate return to VFP command level while form remains active,
* which is only useful under the VFP Development System.
if m.noread_arg									&& caller didn't explicitly suppress READ EVENTS
	return m.demoobjref.ehp_success				&& pick up success flag directly from the live form
endif

* Prepare to use READ EVENTS for modeless cases, since caller didn't explicitly suppress it.
targname = m.demoobjref.ehp_targname			&& pick up indirection target name

* lay down an event loop, preventing immediate return from this main program
READ EVENTS										&& pin down this main program at depth = 1

* return a final success flag result
if type(m.targname + '.ehp_success') = 'L'		&& target error handler is still available
	* return the result by target indirection (since the form is now gone)
	return evaluate(m.targname + '.ehp_success')	&& done - return success flag
endif
return .f.										&& fail, since we have no way to test for success
