* ERRHERR.PRG - ErrHandler core UDF for support of foundation class Error event methods
* By M. Asherman, Copyright (c) 2000 - 2002, SpaceTime Systems
* 11/12/00: adapted from inline code in the ErrHCustom and ErrHForm.Error methods
*	to minimize redundancy and facilitate global refinements (see task erma0009).
*	Also support optional diagnostic tracing logic, 
*	which uses the new ehp_tracelevel property (see erma0004).
* 11/13/00: switch to use of new errhtrac() routine for cleaner tracing output, 
*	which fixes bug due to WAIT WINDOW line length limits.
*	Introduce new diagnostic dialog based on class ErrHDiagDialog - prelim.
* 11/14/00: revise specs to support a return value indicating whether to RETRY,
*	because this must be invoked from the context of the Error event.
*	Also add logic to support the "Ignore" response from dialog.
* 11/17-19/00: phase out use of obsolete errhtrac() by using revised EHM_ShowAppErr().
* 11/21-2/00: Revise logic to support catch-all error handling for cases not anticipated 
*	in method code, i.e. when ehp_traperrors = .F.
* 11/24/00: watch out for delayed Abort cases where the ErrHandler object vanishes. 
* 11/29/00: support new arguments to ehm_showapperr().
* 12/3/00: substitute calls to ehm_copy() for errhcopy().  Avoid direct calls to core UDFs.
*	Add support for the new ehp_on_errx_abort property, for abort-on-exceptional-errors.
* 12/4/00: support the new ehp_ignore_noparam property, for avoiding problems with
*	troublesome too-many-arguments errors.
* 1/12/01: adjust for revised specs as to aborto_arg passed to EHM_ShowAppErr, 
*	and shore up logic to cover all Abort cases during tracing messages (see erma0016).
* 2/12/02: fix logic to return .F. in non-local RETURN TO ... failure cases (erma0024).
* 3/27/02: added fix to avoid errors caused by attemps to RETURN TO ON... commands.
*
* We use a PRG for this logic so it can be shared across multiple base classes,
* which can't have a common ancestor class in VFP.
*
* This function assumes it is called by an ErrHandler Error event,
* which was triggered by an error in that object's code.
*
* Use methods instead invoking their corresponding core UDFs directly, in order to
* support extensions by subclassing.

function errherr
* takes 4 args and returns a flag indicating how the Error event should continue.
* Returns .T. if the Error event should just RETURN, else .F. means it should RETRY.

lparameters errobj_arg, nError, cMethod, nLine		&& takes 4 required arguments
* errobj_arg:		reference to the error-handling object, which should be 
*					derived from one of the MDA foundation error-handling classes.
* nError:			Contains the Visual FoxPro error number.
*					(Same as the corresponding argument to VFP's Error event method.)
* cMethod:			Contains the name of the method that caused the error. 
*					However, if a method calls a user-defined function, 
*					and an error occurs within that function, 
*					cMethod contains the name of the user-defined function, 
*					rather than the name of the method that called the function.
*					(Same as the corresponding argument to VFP's Error event method.)
* nLine:			Contains the line number within the method or user-defined function 
*					that caused the error.
*					(Same as the corresponding argument to VFP's Error event method.)

#include errhandl.h									&& pick up macros for error codes and messages

local msg, retdepth, returnto, periodoff, retval

* perform some minimal argument validation in case of basic usage errors
if vartype(m.errobj_arg) # 'O'						&& not a valid object reference, as required
	* use crude error handling, since we don't have reference to error-handling object
	wait window "ERRHERR Error # " + ltrim(str(ERRH_EC_ERR_INVARG1)) + ': ' ;
	  + ERRH_EM_ERR_INVARG1
	return .T.										&& allow the Error event to RETURN (don't RETRY)
endif

* check for avoidance of problems with troublesome too-many-arguments errors
if m.errobj_arg.ehp_ignore_noparam ;
  and (m.nError = ERRH_VFPERR_NOPARAM ;
  or m.nError = ERRH_VFPERR_NEEDPARAM)				&& ignorable too-many-arguments error
	return .T.										&& Ignore automatically - let Error event RETURN
endif

* capture all potentially useful VFP error information
m.errobj_arg.EHP_VFPErrParam = sys(2018)			&& parameter associated with certain VFP error messages
m.errobj_arg.EHP_VFPErrNo = m.nError				&& unique VFP error #
m.errobj_arg.EHP_VFPErrMethod = m.cMethod			&& name method or UDF in which the error occurred
m.errobj_arg.EHP_VFPErrSourceLineNo = m.nLine		&& line number within method or UDF that caused the error
aerror(m.errobj_arg.EHP_VFPErrArray)				&& obtain array of detailed VFP error info
m.errobj_arg.EHP_VFPErrMsg = message()				&& current error message as a character string
m.errobj_arg.EHP_VFPErrSource = message(1)			&& program line that caused the error, if available
m.errobj_arg.EHP_VFPErrProgDepth = ;
  max(0,program(-1) - 2)							&& numeric value of VFP program level at which error occurred
m.errobj_arg.EHP_VFPErrProgram = ;
  program(m.errobj_arg.EHP_VFPErrProgDepth)			&& name of the program that produced the error
m.errobj_arg.EHP_VFPErrProgPath = ;
  sys(16, m.errobj_arg.EHP_VFPErrProgDepth)			&& file path/name of program that produced the error

* set the generic failure flag
m.errobj_arg.ehp_success = .f.						&& means there was an error

* support optional "indirection" for copying error info into an external errhandler object
m.errobj_arg.ehm_copy(m.errobj_arg.ehp_targobj)		&& no-op if target is null

* see if explicit error trapping is enabled
if m.errobj_arg.ehp_traperrors						&& error trapping is enabled
	* support optional diagnostic tracing
	if m.errobj_arg.ehp_tracelevel > 0 ;
	  and (m.errobj_arg.ehp_tracelevel % ERRH_TRACE_ERRHERR = 0)
		* construct crude diagnostic error message
		msg = 'The Error event trapped an ' ;
		  + iif(m.errobj_arg.ehp_traperrors, 'anticipated', 'unexpected') + ' error.';
		  + chr(13) + "Error args: nError = " + ltrim(str(m.nError)) ;
		  + ", cMethod = " + m.cMethod + ", nLine = " + ltrim(str(m.nLine)) ;
		  + chr(13) + "sys(2018) = " + m.errobj_arg.EHP_VFPErrParam ;
		  + chr(13) + "Error() = " + ltrim(str(Error())) ;
		  + chr(13) + "Message() = " + Message() ;
		  + chr(13) + "Message(1) = " + Message(1)

		* generate tracing output, which may use the diagnostic msg dialog
		* Set abort-to-depth to pop to caller of method or UDF that encountered error.
		m.errobj_arg.EHM_ShowAppErr(0, m.msg, '', 0, ;
		  .t., .t., -3)						&& enable Ignore and Retry buttons
		do case								&& dispatch according to user's response
		  case vartype(m.errobj_arg) # 'O'	&& ErrHandler vanished - presumably Aborted
		  case m.errobj_arg.ehp_diagresponse == ERRH_DIAGRESP_OK	&& default response - OK
			* fall out of this CASE statement and continue with normal error handling logic

		  case m.errobj_arg.ehp_diagresponse == ERRH_DIAGRESP_RETRY
			m.errobj_arg.EHM_Zero()			&& clear error info to retry from a clean slate
			return .F.						&& tells the Error event to RETRY

		  case m.errobj_arg.ehp_diagresponse == ERRH_DIAGRESP_IGNORE
			m.errobj_arg.EHM_Zero()			&& clear error info to treat as successful
			return .T.						&& bypass normal error processing
			* Error event will do a RETURN, which resumes at the line following error line

		  otherwise							&& unknown response - very improbable
		  	* issue a warning message and let normal error handling continue
			m.errobj_arg.EHM_ShowAppErr(0, ;
			  "ERRHERR Warning # " + ltrim(str(ERRH_EC_ERR_INVRESPONSE)) + ': ' ;
			  + ERRH_EM_ERR_INVRESPONSE, .f., .f., -3)
		endcase
	endif

	* Allow the method to continue at the line after the one that triggered Error.
	* This leaves method code responsible for detecting and further handling this error,
	* which avoids putting any further complications into this generic module.
	return .T.								&& tell the calling Error event to RETURN, not RETRY
endif

* Explicit error trapping is not enabled, so use catch-all error handling logic.
* We may need to give some more thought to the special case of errors in the Init method,
* e.g. too many arguments, which triggers an error before we've even had a chance to do
* the basic error handling initializations related to target indirection.
* I'm punting this question for now, since it's pretty obscure.

* Set abort-to-depth to pop to caller of method or UDF that encountered error.
m.errobj_arg.EHM_ShowAppErr(ERRH_EC_ERR_UNEXPECTED, ERRH_EM_ERR_UNEXPECTED ;
  + chr(13) + m.errobj_arg.EHM_VFPErrMsg(), '', 0, ;
  .t., .t., -3)								&& enable Ignore and Retry buttons
do case										&& dispatch according to user's response
  case vartype(m.errobj_arg) # 'O'			&& ErrHandler vanished - presumably Aborted
  case m.errobj_arg.ehp_diagresponse == ERRH_DIAGRESP_OK	&& default response - OK
	* fall out of this CASE statement and continue with default exceptional error handling logic

  case m.errobj_arg.ehp_diagresponse == ERRH_DIAGRESP_RETRY
	* note that Retry doesn't always work, due to VFP inconsistencies
	m.errobj_arg.EHM_Zero()					&& clear error info to retry from a clean slate
	return .F.								&& tells the Error event to RETRY

  case m.errobj_arg.ehp_diagresponse == ERRH_DIAGRESP_IGNORE
	m.errobj_arg.EHM_Zero()					&& clear error info to treat as successful
	return .T.								&& bypass normal error processing
	* Error event will do a RETURN, which resumes at the line following error line

  otherwise									&& unknown response - very improbable
  	* issue a warning message and let exceptional error handling continue
	m.errobj_arg.EHM_ShowAppErr(0, ;
	  "ERRHERR Warning # " + ltrim(str(ERRH_EC_ERR_INVRESPONSE)) + ': ' ;
	  + ERRH_EM_ERR_INVRESPONSE + m.errobj_arg.ehp_diagresponse, .f., .f., -3)
endcase

* Proceed with the default action for exceptional errors
if m.errobj_arg.ehp_on_errx_abort			&& auto-Abort is enabled
	m.errobj_arg.ehm_abort()				&& initiate abnormal form termination
endif

* determine the stack depth to return to
retdepth = max(0,program(-1) - 3)			&& program depth to return to

* skip over any ON... command levels (MDA fix of 3/27/02)
do while m.retdepth > 1 and program(m.retdepth) == 'ON... '	&& prog() returns this for all ON commands
	* this is correct, because ON commands always execute a single command line
	retdepth = m.retdepth - 1				&& pop back an extra level below this ON...
enddo

if m.retdepth < 2							&& called from VFP command window or top-level program
	* return to the highest program level available
	* This still doesn't work as I'd hoped for the case of invocation from VFP command window.
	* Unfortunately, RETURN TO MASTER doesn't return to the VFP command, but it returns 
	* to the program that encountered the error, which is not what I want.
	* It would be interesting to see if executing RETURN TO MASTER from the error event
	* would make any difference; this would require experimentation.
	* Fortunately, this is an obscure problem of no significance to typical runtime apps.
	RETURN TO MASTER .F.					&& this avoids screw-ups that behave like Ignore
endif

* Get program name, checking for imbedded period in method names, which causes a syntax error.
* There seems to be a flaw in VFP's RETURN TO <program> command, in that it can't always return to
* the desired stack level, because the program name may be ambiguous.  This presents a problem
* when we have a recursive program, or even in non-recursive cases if more than one method
* involved in the call stack has the same name.  The consequence of this confusion is that we
* may return to the wrong level of the stack, which could either continue to execute the offending
* method, or could terminate a higher level program invocation than intended.
* Fortunately, neither of these imperfect RETURNs would generally cause any serious problems.
* If we return at too deep a level, the user will see secondary errors that might have been 
* avoided, but probably would be handled cleanly.
* If we return too far back in the stack, the principal of failing with minimum disruption
* may be compromised, but it still beats an even more heavy-handed failure, like Cancel.
* By adding some additional error-checking in logic known to be susceptible to such obscure problems,
* it should be possible to manage despite this known VFP flaw.
returnto = program(m.retdepth)				&& name of program that called the one encountering error
periodoff = rat('.', m.returnto)			&& offset of last period
if m.periodoff > 0							&& period found
	returnto = substr(m.returnto, m.periodoff + 1)	&& strip off the last portion
endif

* return .F. directly to the program that called the one that failed
retval = .F.								&& use a variable instead of literal to avoid syntax error (!)
RETURN TO &returnto m.retval				&& pop the stack back to the caller of method that erred
