* ERRHOERR.PRG - ErrHandler core ON ERROR handler procedure
* By M. Asherman, Copyright (c) 2000 - 2002, SpaceTime Systems
*
* 11/24/00: introduced for support of "catch-all" form-level error handling
*	to cover errors other than those in methods of the form (see erma0005).
*	Adapted logic from similar errherr.prg, which handles events caught via Error methods.
* 11/26/00: force an Activate of the ErrHandler form before invoking Retry,
*	in case diag dialog caused the form to be Deactivated, losing its ON ERROR handler.
* 11/27/00: let diagnostic dialog be responsible for restoring the default ON ERROR handler 
*	instead of doing heavy-handed re-Activate here, which doesn't cover all cases anyhow.
*	Try using RETURN TO ... instead of RETURN to force immediate termination,
*	without resumption of method that encountered abnormal error.
*	(Also see notes in errherr.prg about issues related to use of RETURN TO ...)
* 11/29/00: support new arguments to ehm_showapperr().
*	Respect the new ehp_on_err_abort option, so that Abort isn't mandatory
*	(Abort need not even be the default).
* 12/3/00: beef up validation logic, esp. in case we failed to restore the 
*	original ON ERROR handler.  Forget about propagating extraordinary errors
*	via VFP's ERROR command, because VFP doesn't support nested error handling.
*	Support additional arguments to pass the values of message() and sys(2018)
*	instead of relying on picking these up internally, because the use of 
*	VFP's TYPE() function can alter the original message parameter if the 
*	initial validation fails (e.g. incorrectly yielding sys(2018) = "ACTIVEFORM").
* 12/4/00: support the new ehp_ignore_noparam property, for avoiding problems with
*	troublesome too-many-arguments errors.
* 1/12/01: adjust for revised specs as to aborto_arg passed to EHM_ShowAppErr, 
*	and shore up logic to cover all Abort cases during tracing messages (see erma0016).
* 2/12/02: fix logic to return .F. in non-local RETURN TO ... failure cases (erma0024).
* 3/27/02: added fix to avoid errors caused by attemps to RETURN TO ON... commands.
* 5/28/02: added logic to support a new global error handling capability.
*
* For the sake of consistency, support arguments like those passed to VFP's 
* Error event method, although we really could manage with only the line # argument.

lparameters nError, cMethod, nLine, cMessage, cParam		&& expects 5 arguments
* nError:			Contains the Visual FoxPro error number.
*					(Same as the corresponding argument to VFP's Error event method.)
* cMethod:			Contains the name of the method or program that caused the error. 
*					If a method calls a user-defined function, 
*					and an error occurs within that function, 
*					cMethod contains the name of the user-defined function, 
*					rather than the name of the method that called the function.
*					(Similar to corresponding argument for VFP's Error event method.)
* nLine:			Contains the line number within the method or user-defined function 
*					that caused the error.
*					(Same as the corresponding argument to VFP's Error event method.)
* cMessage:			VFP error message text, as returned by MESSAGE().
* cParam:			VFP error message parameter, as returned by SYS(2018).


#include errhandl.h							&& pick up macros for error codes and messages

local waitmsg, errnum, errobj, retdepth, returnto, periodoff, retval

* get error # from argument or error() if arg non-numeric
errnum = iif(vartype(m.nError) = 'N', m.nError, ERROR())

* try to get object reference to the active ErrHandler-based form.
if type('_screen.activeform.EHP_on_err_handle') = 'L' ;
  and _screen.activeform.EHP_on_err_handle	&& active form wants to handle these errors
	errobj = _screen.activeform				&& -> active form
else										&& not handled by active form
	* check for a global error handler object
	if vartype(m.ERRH_MVGLOBAL) = 'O'		&& reserved variable -> non-null object
		errobj = m.ERRH_MVGLOBAL			&& -> global error handler
	else									&& there is no available ErrHandler object
		* This error really should never happen, because this handler is established by
		* an ErrHandler-based form's Activate event, and this handler should be disabled
		* whenever the form is deactivated.  This extra check is just to play it safe.

		* Use crude error reporting, since we don't have reference to error-handling object
		waitmsg = "ERRHOERR Error # " + ltrim(str(ERRH_EC_ONERR_NOFORM)) + ': ' ;
		  + ERRH_EM_ONERR_NOFORM ;
		  + chr(13) + "nError = " + iif(vartype(m.nError) = 'N', ;
		  ltrim(str(m.nError)), '(type = "' + vartype(m.nError) + '")') ;
		  + ", cMethod = " + iif(vartype(m.cMethod) = 'C', ;
		  m.cMethod, '(type = "' + vartype(m.cMethod) + '")') ;
		  + ", nLine = " + iif(vartype(m.nLine) = 'N', ;
		  ltrim(str(m.nLine)), '(type = "' + vartype(m.nLine) + '")') ;
		  + chr(13) + "cMessage = " + iif(vartype(m.cMessage) = 'C', ;
		  m.cMessage, '(type = "' + vartype(m.cMessage) + '")') ;
		  + chr(13) + "cParam = " + iif(vartype(m.cParam) = 'C', ;
		  m.cParam, '(type = "' + vartype(m.cParam) + '")')
		if len(m.waitmsg) > ERRH_MAXWAITWINDLEN			&& exceeds limit for WAIT WINDOW
			waitmsg = substr(m.waitmsg, 1, ERRH_MAXWAITWINDLEN - 3) + "..."
		endif
		WAIT WINDOW m.waitmsg				&& display warning

		* get out here as fast as possible
		ON ERROR							&& revert to VFP's default handler
		return								&& resume at line after the error
	endif
endif

* check for avoidance of problems with troublesome too-many-arguments errors
if m.errobj.ehp_ignore_noparam ;
  and (m.errnum = ERRH_VFPERR_NOPARAM ;
  or m.errnum = ERRH_VFPERR_NEEDPARAM)		&& ignorable too-many-arguments error
	return .T.								&& Ignore automatically - let Error event RETURN
endif

* capture all potentially useful VFP error information
m.errobj.EHP_VFPErrParam = iif(vartype(m.cParam) = 'C', ;
  m.cParam, sys(2018))						&& parameter associated with certain VFP error messages
m.errobj.EHP_VFPErrNo = m.errnum			&& unique VFP error #
m.errobj.EHP_VFPErrMethod = iif(vartype(m.cMethod) = 'C', ;
  m.cMethod, PROGRAM())						&& name method or UDF in which the error occurred
m.errobj.EHP_VFPErrSourceLineNo = iif(vartype(m.nLine) = 'N', ;
  m.nLine, 0)								&& line # within prog that caused error (0 = unknown)
aerror(m.errobj.EHP_VFPErrArray)			&& obtain array of detailed VFP error info
m.errobj.EHP_VFPErrMsg = iif(vartype(m.cMessage) = 'C', ;
  m.cMessage, message())					&& current error message as a character string
m.errobj.EHP_VFPErrSource = message(1)		&& program line that caused the error, if available
m.errobj.EHP_VFPErrProgDepth = ;
  max(0,program(-1) - 2)					&& numeric value of VFP program level at which error occurred
m.errobj.EHP_VFPErrProgram = ;
  program(m.errobj.EHP_VFPErrProgDepth)		&& name of the program that produced the error
m.errobj.EHP_VFPErrProgPath = ;
  sys(16, m.errobj.EHP_VFPErrProgDepth)		&& file path/name of program that produced the error

* invoke catch-all error reporting for abnormal errors while form is active
* Set abort-to-depth to pop to caller of method or UDF that encountered error.
m.errobj.EHM_ShowAppErr(ERRH_EC_ONERR_ABNORMAL, ERRH_EM_ONERR_ABNORMAL ;
  + chr(13) + m.errobj.EHM_VFPErrMsg(), '', 0, ;
  .t., .t., -3)								&& enable Ignore and Retry buttons

* check for replies to the diagnostic message dialog, if it was used
do case										&& dispatch according to user's response
  case vartype(m.errobj) # 'O'				&& ErrHandler vanished - presumably Aborted
  	* maybe could achieve more immediate abort with RETURN TO ... - not yet tried
	RETURN									&& allow Abort in progress to complete

  case m.errobj.ehp_diagresponse == ERRH_DIAGRESP_OK	&& default response - OK
	* fall out of this CASE statement and continue with default abnormal error handling logic

  case m.errobj.ehp_diagresponse == ERRH_DIAGRESP_RETRY
	* note that Retry doesn't always work, due to VFP inconsistencies
	m.errobj.EHM_Zero()						&& clear error info to retry from a clean slate
	RETRY									&& re-attempt the command line that failed

  case m.errobj.ehp_diagresponse == ERRH_DIAGRESP_IGNORE
	m.errobj.EHM_Zero()						&& clear error info to treat as successful
	RETURN									&& bypass abnormal error processing
	* resume at the line following error line, as if nothing went wrong

  otherwise									&& unknown response - very improbable
  	* issue a warning message and let default abnormal error handling continue
	m.errobj.EHM_ShowAppErr(0, ;
	  "ERRHOERR Warning # " + ltrim(str(ERRH_EC_ONERR_INVRESPONSE)) + ': ' ;
	  + ERRH_EM_ONERR_INVRESPONSE + m.errobj.ehp_diagresponse, .f., .f., -3)
endcase

* Proceed with the default action for abnormal errors
if m.errobj.ehp_on_err_abort				&& auto-Abort is enabled
	m.errobj.ehm_abort()					&& initiate abnormal form termination
endif

* Try to RETURN in such a way as to avoid resuming in method that encountered error.
retdepth = max(0,program(-1) - 3)			&& program depth to return to

* skip over any ON... command levels (MDA fix of 3/27/02)
do while m.retdepth > 1 and program(m.retdepth) == 'ON... '	&& prog() returns this for all ON commands
	* this is correct, because ON commands always execute a single command line
	retdepth = m.retdepth - 1				&& pop back an extra level below this ON...
enddo

* avoid ambiguous return location whenever possible
if m.retdepth < 2							&& called from VFP command window or top-level program
	RETURN TO MASTER .F.					&& return to the highest program level available
endif

* Get program name, checking for imbedded period in method names, which causes a syntax error.
* (See notes in errherr.prg about flaws in VFP's RETURN TO <program> command.)
returnto = program(m.retdepth)				&& name of program that called the one encountering error
periodoff = rat('.', m.returnto)			&& offset of last period
if m.periodoff > 0							&& period found
	returnto = substr(m.returnto, m.periodoff + 1)	&& keep only the last portion
endif

* return .F. directly to the program that called the one that failed
retval = .F.								&& use a variable instead of literal to avoid syntax error (!)
RETURN TO &returnto m.retval				&& pop the stack back to the caller of method that erred
