* ERRHTARG.PRG - UDF for initializing indirection target in MDA generic error-handling classes
* By M. Asherman, Copyright (c) 1999, SpaceTime Systems
* 11/24/99: adapted from logic previously located in errhfrom.init - prelim.
* 11/26/99: cleaned up preliminary internal error reporting in this function,
*	to make proper use of approved generic procedures, for gosh sake!
*	Also revised logic to force initial call to errhzero(), to observe the
*	convention that this always be done by functions supporting failure detection.
* 11/10/00: support the new optional targob_arg, for even more flexibility in
*	establishing target indirection, in order to cover Init-time errors (see erma0008).
*	Also beef up checks for usage errors for the benefit of developers, since this
*	is the first core procedure likely to be encountered.
* 12/3/00: Avoid direct calls to core UDFs by substituting calls to corresponding methods.
* 12/28/00: pick up default diagnostic/tracing properties via target indirection (erma0010).
* 1/12/01: for unexpected errors instead of unconditionally returning .T. (see erma0016).
*
* This function deals with initialization of an error handler's ehp_targobj property,
* which is the object reference to an additional error handler to be maintained,
* e.g. for returning detailed error info from a modal form or for Init-time errors.
*
* This function supports explicit error-handling for its own internal errors,
* so it always clears and/or sets the primary object's error properties,
* and it returns a success flag result.  Used internally in foundation Init methods.

function errhtarg
* takes 1 - 5 arguments returns a success flag

lparameters errobj_arg, targob_arg, targnm_arg, tclass_arg, tarlib_arg	&& takes 1 - 5 arguments
* errobj_arg:		reference to the error-handling object, which should be 
*					derived from one of the MDA foundation error-handling classes.
* targob_arg:		optional object reference to the target ErrHandler to be used.
*					If supplied, ignore the remaining optional arguments.
*					If NULL, empty, or omitted, use the remaining arguments to determine
*					the target ErrHandler requirements and perform any related initializations.
* targnm_arg:		name of the indirection target, which may be either a memory variable
*					or a member of an existing object (e.g. _screen).
*					If this target error-handler doesn't already exist, it is created
*					via newobject(), and the reference is retained in the primary errobj_arg.
*					A PUBLIC mvar is allocated and initialized to refer to the target, 
*					if no such mvar already exists.
* tclass_arg:		Class name of the target error handler to create, if necessary.
*					This will be passed to the newobject() function or method, as appropriate.
* tarlib_arg:		Class library name of the target error handler to create, if necessary.
*					This will be passed to the newobject() function or method, as appropriate.

#include errhandl.h				&& pick up macros for error codes and messages

local ismvar, parentname, parentobj, membname

* validate the required first argument, in case of a basic usage error
if type('m.errobj_arg.ehp_apperrcode') # 'N'	&& not a valid ErrHandler object
	* use crude error handling, since we don't have reference to error-handling object
	wait window "ERRHTARG Error # " + ltrim(str(ERRH_EC_TARG_NOOBJ)) + ': ' ;
	  + ERRH_EM_TARG_NOOBJ
	return .f.									&& fail
endif

* clear any previous residual error properties to indicate success of this call
m.errobj_arg.EHM_Zero()							&& start from normal no-error state

* check for optional explicit object reference to the target ErrHandler to be used
if vartype(m.targob_arg) = 'O' and not isnull(m.targob_arg)		&& non NULL object ref
	* confirm that this appears to be a valid ErrHandler
	if type('m.targob_arg.ehp_apperrcode') # 'N'	&& not a valid ErrHandler object
		* use generic error reporting with macro-defined error code and message
		m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_INVTARG, ERRH_EM_TARG_INVTARG)
		return .f.								&& fail
	endif
	
	* Check for direct self-reference errors, but note that we are punting the
	* more general issue of transitively circular cases.
	if m.errobj_arg = m.targob_arg				&& self-reference - the trivial circular case
		* use generic error reporting with macro-defined error code and message
		m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_ISSELF, ERRH_EM_TARG_ISSELF)
		return .f.								&& fail
	endif
	
	* hold on to the caller-supplied target
	m.errobj_arg.ehp_targobj = m.targob_arg		&& establish target for indirection
	m.errobj_arg.ehm_copydiags(m.targob_arg)	&& copy diag settings FROM the target
	m.errobj_arg.EHM_Zero()						&& propagate zeroing TO the new target
	return m.errobj_arg.ehp_success				&& success unless unexpected error
else											&& not an object reference
	if vartype(m.targob_arg) # 'O' and not empty(m.targob_arg)	&& invalid non-empty value
		* use generic error reporting with macro-defined error code and message
		m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_INVTARGTYP, ERRH_EM_TARG_INVTARGTYP)
		return .f.								&& fail
	endif
endif

* check for automatic indirection to a default errhandler object
if empty(m.targnm_arg)							&& no target name 
	* Not sure if this should be a no-op or should release any prior target;
	* it's a toss-up, so I opted to clear prior target for now.
	m.errobj_arg.ehp_targobj = .null.			&& discard any prior target indirection
	return m.errobj_arg.ehp_success				&& success unless unexpected error
endif

* there is a non-empty target name - see if it already exists and is usable as-is
if type(m.targnm_arg + '.ehp_apperrcode') = 'N'			&& looks like a valid errhandler
	m.errobj_arg.ehp_targobj = evaluate(m.targnm_arg)	&& -> existing error handler
	m.errobj_arg.ehm_copydiags(m.errobj_arg.ehp_targobj)	&& copy settings FROM the target
	m.errobj_arg.EHM_Zero()						&& propagate error properties to new target
	return m.errobj_arg.ehp_success				&& success unless unexpected error
endif

* the named target error handler doesn't yet exist, so prepare to create it
do case										&& sort out name cases
  case not '.' $ m.targnm_arg				&& name contains no period
  	ismvar = .t.							&& assume this is a memory variable name

  case at('m.', lower(m.targnm_arg)) = 1 and occurs('.', m.targnm_arg) = 1	&& begins with "m."
  	ismvar = .t.							&& assume this is a memory variable name

  otherwise									&& not a memory variable
  	ismvar = .f.							&& assume it's a member object name
endcase

* Handle mvar cases first
if m.ismvar									&& target name could only be a memory variable
	* be sure the mvar is defined
	if type(m.targnm_arg) = 'U'				&& no such mvar is yet defined
		m.errobj_arg.ehp_traperrors = .t.	&& enable error trapping around PUBLIC declaration
		public (m.targnm_arg)				&& allocate specified PUBLIC mvar
		m.errobj_arg.ehp_traperrors = .f.	&& revert to general default ON ERROR heandling
		if not m.errobj_arg.ehp_success or type(m.targnm_arg) = 'U'		&& PUBLIC decl failed
			* use generic error reporting with macro-defined error code and message
			m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_NOPUB, ;
			  ERRH_EM_TARG_NOPUB + m.targnm_arg ;
			  + chr(13) + m.errobj_arg.EHM_VFPErrMsg())
			return .f.						&& failure
		endif
		store .null. to (m.targnm_arg)		&& initialize to null object ref
	endif
	
	* allocate and initialize mvar, if there is no conflicting use 
	if isnull(evaluate(m.targnm_arg))		&& target mvar = .NULL.
		* create a new errhandler object and save reference in specified mvar
		m.errobj_arg.ehp_traperrors = .t.	&& enable error trapping around newobject()
		store newobject(m.tclass_arg, m.tarlib_arg) to (m.targnm_arg)	&& allocate errhandler
		m.errobj_arg.ehp_traperrors = .f.	&& revert to general default ON ERROR heandling
		m.errobj_arg.ehp_targobj = evaluate(m.targnm_arg)	&& hold on to new errhandler object ref.
		if not m.errobj_arg.ehp_success or isnull(m.errobj_arg.ehp_targobj)	&& newobject() failed
			* use generic error reporting with macro-defined error code and message
			m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_NONEW, ;
			  ERRH_EM_TARG_NONEW + m.targnm_arg ;
			  + chr(13) + "Class: " + m.tclass_arg + ", Lib: " + m.tarlib_arg ;
			  + chr(13) + m.errobj_arg.EHM_VFPErrMsg())
			return .f.						&& failure
		endif
		m.errobj_arg.ehm_copydiags(m.errobj_arg.ehp_targobj)	&& copy settings FROM the target
		m.errobj_arg.EHM_Zero()				&& force propagation error properties to new target
		return m.errobj_arg.ehp_success		&& success unless unexpected error
	endif
	
	* use generic error reporting with macro-defined error code and message
	m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_INUSE, ;
	  ERRH_EM_TARG_INUSE + m.targnm_arg)
	return .f.								&& fail
endif

* target name specifies a member object, not an mvar
parentname = substr(m.targnm_arg, 1, rat('.', m.targnm_arg) - 1)	&& parent object name
if type(m.parentname) # 'O' or isnull(evaluate(m.parentname))		&& parent doesn't exist
	* use generic error reporting with macro-defined error code and message
	m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_NOPA, ;
	  ERRH_EM_TARG_NOPA + m.parentname)
	return .f.								&& fail
endif

* parent object exists - check that target member is not yet present
if type(m.targnm_arg) # 'U'					&& target member exists, but it's not usable
	* use generic error reporting with macro-defined error code and message
	m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_INVMEMB, ;
	  ERRH_EM_TARG_INVMEMB + m.targnm_arg)
	return .f.								&& fail
endif

* use ...newobject to create new member object
parentobj = evaluate(m.parentname)			&& -> parent object
membname = substr(m.targnm_arg, rat('.', m.targnm_arg) + 1)	&& member name to be used for errhandler
m.errobj_arg.ehp_traperrors = .t.			&& enable error trapping around newobject()
m.parentobj.newobject(m.membname, m.tclass_arg, m.tarlib_arg)		&& create new member
m.errobj_arg.ehp_traperrors = .f.			&& revert to general default ON ERROR heandling
if not m.errobj_arg.ehp_success or type(m.targnm_arg) # 'O'			&& failed to allocate member
	* use generic error reporting with macro-defined error code and message
	m.errobj_arg.EHM_ShowAppErr(ERRH_EC_TARG_NONEWM, ;
	  ERRH_EM_TARG_NONEWM + m.targnm_arg ;
	  + chr(13) + "Class: " + m.tclass_arg + ", Lib: " + m.tarlib_arg ;
	  + chr(13) + m.errobj_arg.EHM_VFPErrMsg())
	return .f.								&& failure
endif

* created target errhandler member successfully
m.errobj_arg.ehp_targobj = evaluate(m.targnm_arg)	&& save ref. to external error handler
m.errobj_arg.ehm_copydiags(m.errobj_arg.ehp_targobj)	&& copy settings FROM the target
m.errobj_arg.EHM_Zero()						&& force propagation error properties to new target
return m.errobj_arg.ehp_success				&& success unless unexpected error
